/*
	Dinamise - Dinamise Is Not A MIdi SEquencer

	Advanced music sequencer for GBA
	Copyright (C) 2005  Avelino Herrera Morales

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "gba.h"
#include "font.h"

#define  PRESSED(k)  (!(REG_KEYPAD & (k)))

int __gba_multiboot;

/* modo grfico */
#define BG0_ENABLE      0x100
#define BG1_ENABLE      0x200
#define BG2_ENABLE      0x400
#define BG3_ENABLE      0x800
#define OBJ_ENABLE      0x1000
#define WIN0_ENABLE     0x2000
#define WIN1_ENABLE     0x4000
#define WINOBJ_ENABLE   0x8000

#define BACK_BUFFER     0x10

#define MODE0           0x0
#define MODE1           0x1
#define MODE2           0x2
#define MODE3           0x3
#define MODE4           0x4
#define MODE5           0x5

#define  SCREEN_BUFFER  ((u16 *) 0x6000000)
#define  TILE_BUFFER    ((u16 *) 0x6004000)
#define  BG_PALETTE     ((u16 *) 0x5000000)

#define RGB(r,g,b)  ((r) + ((g) << 5) + ((b) << 10))

#define  ERASE_SCREEN  memset(SCREEN_BUFFER, 0, 0x800)

u16 standard_color_palette[16] = {
	RGB(0, 0, 0),
	RGB(11, 11, 11),
	RGB(22, 4, 4),
	RGB(5, 9, 27),
	RGB(4, 13, 3),
	RGB(16, 9, 3),
	RGB(16, 5, 24),
	RGB(20, 20, 20),
	RGB(16, 25, 15),
	RGB(20, 22, 31),
	RGB(5, 26, 26),
	RGB(31, 18, 6),
	RGB(31, 30, 6),
	RGB(29, 28, 23),
	RGB(31, 26, 30),
	RGB(31, 31, 31)
};

#define  BLACK        0
#define  DARK_GRAY    1
#define  RED          2
#define  BLUE         3
#define  GREEN        4
#define  BROWN        5
#define  PURPLE       6
#define  LIGHT_GRAY   7
#define  LIGHT_GREEN  8
#define  LIGHT_BLUE   9
#define  CYAN         10
#define  ORANGE       11
#define  YELLOW       12
#define  TAN          13
#define  PINK         14
#define  WHITE        15

#define  NUM_NOTES  12
char note_string[NUM_NOTES][3] = {
	"c-", "c#", "d-", "d#", "e-", "f-", "f#", "g-", "g#", "a-", "a#", "b-"
};

#define  NUM_OCTAVES  10
char octave_character[] = "0123456789A";

#define  PATTERN_ROW_TYPE_EMPTY       0
#define  PATTERN_ROW_TYPE_NOTE_ON     1
#define  PATTERN_ROW_TYPE_NOTE_OFF    2
#define  PATTERN_ROW_TYPE_CONTROLLER  3

typedef struct {
	unsigned char type;
	unsigned char col0;
	unsigned char col1;
} pattern_row;   /* 4 bytes */

#define  MAX_ROWS          64
#define  MAX_VISIBLE_ROWS  11

typedef struct {
	pattern_row data[MAX_ROWS];
} pattern_track;  /* 256 bytes */

#define  NUM_TRACKS          14
#define  MAX_VISIBLE_TRACKS  3

typedef struct {
	u16 num_rows;
	u16 hilight_step;
	pattern_track tracks[NUM_TRACKS];
} pattern; /* 4100 bytes */

#define  MAX_PATTERNS        16
#define  MAX_ORDERS          100
#define  MAX_VISIBLE_ORDERS  8
#define  MIN_BPM             20
#define  MAX_BPM             300
#define  DEFAULT_BPM         120

typedef struct {
	char signature[8];                    /* "DINAMISE" */
	pattern patterns[MAX_PATTERNS];
	char order[MAX_ORDERS];
	unsigned char midi_channel[NUM_TRACKS];
	u16 bpm;
} song;

song *current_song;

u16 current_pattern_index = 0;
char aux_str[80];
char a_field_str[10];
char b_field_str[10];
char c_field_str[10];
u16 cursor_row = 0;
u16 cursor_track = 0;
u16 display_track_offset = 0;
u16 display_row_offset = 0;
u16 cursor_order = 0;
u16 display_order_offset = 0;

#define  FILE_OPTION_NEW   0
#define  FILE_OPTION_LOAD  1
#define  FILE_OPTION_SAVE  2
#define  NUM_FILE_OPTIONS  3
u16 cursor_option = FILE_OPTION_NEW;

typedef struct {
	char label[10];
	void (*callback)(void);
} file_option;

void file_option_new(void);
void file_option_load(void);
void file_option_save(void);

file_option file_options[] = {
	{"NEW", file_option_new},
	{"LOAD", file_option_load},
	{"SAVE", file_option_save}
};

#define  PATTERN_ROW_FIELD_NOTE_CONTROLLER  0
#define  PATTERN_ROW_FIELD_OCTAVE_X         1
#define  PATTERN_ROW_FIELD_VELOCITY_VALUE   2

u16 cursor_field = PATTERN_ROW_FIELD_NOTE_CONTROLLER;

#define  STATUS_EDITING_PATTERN  0
#define  STATUS_EDITING_ORDER    1
#define  STATUS_EDITING_BPM      2
#define  STATUS_FILE             3
#define  STATUS_PLAYING          4

u16 status = STATUS_EDITING_PATTERN;

#define  EDITING_PATTERN_STATUS_PAT       0
#define  EDITING_PATTERN_STATUS_SIZE      1
#define  EDITING_PATTERN_STATUS_HLSTEP    2
#define  EDITING_PATTERN_STATUS_CHANNELS  3
#define  EDITING_PATTERN_STATUS_PATTERN   4

u16 editing_pattern_status = EDITING_PATTERN_STATUS_PAT;

struct {
	u8 *data;
	u8 data_size;
	s8 next_bit;
	u8 byte;
} midi_send_struct;

// SO = 3.3 v   (always 1)  (SO is SI in the slave end of the cable)
// SC = MIDI stream         (SC is SC too in the slave end of the cable)
#define  REG_R_0_VALUE  0x80F8
#define  REG_R_1_VALUE  0x80F9

char status_string_editing[] = "status: editing";
char status_string_playing[] = "status: playing";
char status_string_file[] = "status: file";
char status_string_new_ok[] = "status: new ok";
char status_string_load_ok[] = "status: load ok";
char status_string_load_error[] = "status: no song in SRAM";
char status_string_save_ok[] = "status: save ok";
char *status_string = status_string_editing;

#define  MIDI_CLOCKS_PER_MINUTE  ((u32)31242L * (u32)60L)
#define  MIDI_CLOCKS_BY_BPM(b)   (MIDI_CLOCKS_PER_MINUTE / (((u32) (b)) * 4))

u32 current_bpm_midi_clocks = MIDI_CLOCKS_BY_BPM(DEFAULT_BPM);
u32 bpm_midi_clocks_counter = MIDI_CLOCKS_BY_BPM(DEFAULT_BPM);
u16 trig_row = 0;

u8 *sram = (u8 *) 0x0E000000;

/* playing status */
pattern *playing_pattern;
u16 playing_row_index;
u16 playing_order_index;
u8 midi_buffer[NUM_TRACKS * 3];
u8 midi_buffer_size = 0;
u8 track_last_note_on[NUM_TRACKS];


void InterruptProcess(void) __attribute__ ((section(".iwram")));   // the ISR must be declared in this way

 
void InterruptProcess(void) {
	u8 bit;

	/* midi */
	if (midi_send_struct.byte >= midi_send_struct.data_size)
		bit = 1;
	else if (midi_send_struct.next_bit == -1) {
		midi_send_struct.next_bit = 0;
		bit = 0;
	}
	else if (midi_send_struct.next_bit > 7) {
		midi_send_struct.byte++;
		midi_send_struct.next_bit = -1;
		bit = 1;
	}
	else {
		bit = 1 & (midi_send_struct.data[midi_send_struct.byte] >> midi_send_struct.next_bit);
		midi_send_struct.next_bit++;
	}
	REG_R = (bit) ? REG_R_1_VALUE : REG_R_0_VALUE;
	/* row counter */
	if (status == STATUS_PLAYING) {
		if (bpm_midi_clocks_counter > 0)
			bpm_midi_clocks_counter--;
		else {
			bpm_midi_clocks_counter = current_bpm_midi_clocks;
			trig_row = 1;
		}
	}
	/* clear interrupt flag */
	REG_IF |= REG_IF;
}


void timer_0_start(void) {
	midi_send_struct.data_size = 0;
	midi_send_struct.byte = -1;
	// stop timer 0
	REG_TM0CNT_H = 0;
	// enable irq for timer 0
	REG_IE = 0x0008;
	REG_IME = 1;
	// load timer 0
	// 65536 - 64999 = 537
	// (16 * 1024 * 1024) / 537 = 31242 bps
	// MIDI is 31250 bsp at +/-1% tolerance (-1% => 30937, +1% => 31562)
	// so 31242 bps is a valid bps value
	REG_TM0CNT_L = 0xFDE7;   // 64999
	REG_TM0CNT_H = 0x80 | 0x40;   // start + irq + f (16MHz)
}


void timer_0_stop(void) {
	// stop timer 0
	REG_TM0CNT_H = 0;
}


u8 midi_send(u8 *data, u8 size) {
	if (midi_send_struct.byte < midi_send_struct.data_size)
		return 0;
	midi_send_struct.data = data;
	midi_send_struct.data_size = size;
	midi_send_struct.next_bit = -1;
	midi_send_struct.byte = 0;
	return 1;
}


void start_playing(void) {
	u8 pattern_index = current_song->order[0];
	playing_order_index = 0;
	if (pattern_index < MAX_PATTERNS) {
		playing_pattern = &(current_song->patterns[pattern_index]);
		memset(track_last_note_on, 0, NUM_TRACKS);
		playing_row_index = 0;
		current_bpm_midi_clocks = MIDI_CLOCKS_BY_BPM(current_song->bpm);
		bpm_midi_clocks_counter = current_bpm_midi_clocks;
	}
	else
		playing_pattern = NULL;
}


void stop_playing(void) {
	playing_pattern = NULL;
	status_string = status_string_editing;
	status = STATUS_EDITING_PATTERN;
}


void poll_playing(void) {
	u16 i, next_pattern_index;

	if ((trig_row == 0) || (playing_pattern == NULL))
		return;
	midi_buffer_size = 0;
	for (i = 0; i < NUM_TRACKS; i++) {
		pattern_row *row = &(playing_pattern->tracks[i].data[playing_row_index]);
		if (row->type == PATTERN_ROW_TYPE_NOTE_ON) {
			midi_buffer[midi_buffer_size++] = 0x90 + (current_song->midi_channel[i] - 1);
			midi_buffer[midi_buffer_size++] = row->col0;
			midi_buffer[midi_buffer_size++] = row->col1;
			track_last_note_on[i] = row->col0;
		}
		else if (row->type == PATTERN_ROW_TYPE_NOTE_OFF) {
			midi_buffer[midi_buffer_size++] = 0x80 + (current_song->midi_channel[i] - 1);
			midi_buffer[midi_buffer_size++] = track_last_note_on[i];
			midi_buffer[midi_buffer_size++] = 64;
		}
		else if (row->type == PATTERN_ROW_TYPE_CONTROLLER) {
			midi_buffer[midi_buffer_size++] = 0xB0 + (current_song->midi_channel[i] - 1);
			midi_buffer[midi_buffer_size++] = row->col0;
			midi_buffer[midi_buffer_size++] = row->col1;
		}
	}
	midi_send(midi_buffer, midi_buffer_size);
	/* goto next row */
	if (playing_row_index < (playing_pattern->num_rows - 1))
		playing_row_index++;
	else {
		if (playing_order_index < (MAX_ORDERS - 1)) {
			playing_order_index++;
			playing_row_index = 0;
			next_pattern_index = current_song->order[playing_order_index];
			if (next_pattern_index != 0xFF)
				playing_pattern = &(current_song->patterns[next_pattern_index]);
			else
				stop_playing();
		}
		else
			stop_playing();
	}
	trig_row = 0;
}


#define  WAIT_RELEASE(button)  while (PRESSED(button))


u16 key_pressed(u16 k) {
	if (PRESSED(k)) {
		WAIT_RELEASE(k);
		return 1;
	}
	else
		return 0;
}


void dump_bitmap_font_to_tile_buffer(unsigned char *font, unsigned char foreground_color, unsigned char background_color) {
	int i, bit;
	unsigned char *from;
	u16 *to, aux;
	unsigned char part;

	from = font;
	to = TILE_BUFFER;
	for (i = 0; i < FONT_SIZE; i++, from++) {
		/* each byte (row) of a character data are 4 bytes of tile data */
		aux = 0;
		part = 0;
		for (bit = 7; bit >= 0; bit--) {
			u16 color = background_color;
			if ((*from >> bit) & 1)
				color = foreground_color;
			if (part == 0)
				aux |= color;
			else if (part == 1)
				aux |= (color << 4);
			else if (part == 2)
				aux |= (color << 8);
			else if (part == 3)
				aux |= (color << 12);
			if (part == 3) {
				*to = aux;
				to++;
				part = 0;
				aux = 0;
			}
			else
				part++;
		}
	}
}

#define  LIGHT_GRAY_OVER_BLACK   0
#define  WHITE_OVER_BLACK        1
#define  LIGHT_GRAY_OVER_BLUE    2
#define  WHITE_OVER_BLUE         3
#define  LIGHT_GREEN_OVER_BLACK  4
#define  YELLOW_OVER_BLACK       5
#define  YELLOW_OVER_BLUE        6

void define_palettes(void) {
	u16 *p;

	/* palette 0 */
	p = BG_PALETTE;
	p[0] = standard_color_palette[BLACK];
	p[1] = standard_color_palette[BLACK];
	p[2] = standard_color_palette[LIGHT_GRAY];
	/* palette 1 */
	p += 16;
	p[0] = standard_color_palette[BLACK];
	p[1] = standard_color_palette[BLACK];
	p[2] = standard_color_palette[WHITE];
	/* palette 2 */
	p += 16;
	p[0] = standard_color_palette[BLACK];
	p[1] = standard_color_palette[BLUE];
	p[2] = standard_color_palette[LIGHT_GRAY];
	/* palette 3 */
	p += 16;
	p[0] = standard_color_palette[BLACK];
	p[1] = standard_color_palette[BLUE];
	p[2] = standard_color_palette[WHITE];
	/* palette 4 */
	p += 16;
	p[0] = standard_color_palette[BLACK];
	p[1] = standard_color_palette[BLACK];
	p[2] = standard_color_palette[LIGHT_GREEN];
	/* palette 5 */
	p += 16;
	p[0] = standard_color_palette[BLACK];
	p[1] = standard_color_palette[BLACK];
	p[2] = standard_color_palette[YELLOW];
	/* palette 5 */
	p += 16;
	p[0] = standard_color_palette[BLACK];
	p[1] = standard_color_palette[BLUE];
	p[2] = standard_color_palette[YELLOW];
}


void put_character(char c, u16 x, u16 y, unsigned char palette) {
	int offset;
	u16 value;

	if (c < 32)
		c = 32;
	value = ((u16) (c - 32)) | (((u16) palette) << 12);
	offset = (y * 32) + x;
	SCREEN_BUFFER[offset] = value;
}


void put_string(char *s, u16 x, u16 y, unsigned char palette) {
	while (*s != 0) {
		put_character(*s, x, y, palette);
		x++;
		s++;
	}
}


void wait_for_vsync(void) {
	while (((volatile u16) REG_VCOUNT) != 160)
		;
}


#define  GET_NOTE_INDEX(midi_note)    ((midi_note) % NUM_NOTES)
#define  GET_OCTAVE_INDEX(midi_note)  ((midi_note) / NUM_NOTES)


void draw_pattern_row(pattern_row *r, u16 x, u16 y, u16 last, u16 border_palette, u16 internal_palette, u16 cursor_at_this_row, u16 hilighted_row) {
	u16 a_field_palette, b_field_palette, c_field_palette;

	if (r == NULL)
		return;
	a_field_palette = internal_palette;
	b_field_palette = internal_palette;
	c_field_palette = internal_palette;
	if (cursor_at_this_row && (status == STATUS_EDITING_PATTERN) && (editing_pattern_status == EDITING_PATTERN_STATUS_PATTERN)) {
		u16 palette_to_assign = (hilighted_row) ? WHITE_OVER_BLUE : LIGHT_GRAY_OVER_BLUE;
		if (cursor_field == PATTERN_ROW_FIELD_NOTE_CONTROLLER)
			a_field_palette = palette_to_assign;
		else if (cursor_field == PATTERN_ROW_FIELD_OCTAVE_X)
			b_field_palette = palette_to_assign;
		else if (cursor_field == PATTERN_ROW_FIELD_VELOCITY_VALUE)
			c_field_palette = palette_to_assign;
	}
	put_character('|', x, y, border_palette);
	if (r->type == PATTERN_ROW_TYPE_EMPTY) {
		strcpy(a_field_str, "--");
		strcpy(b_field_str, "-");
		strcpy(c_field_str, "--");
	}
	else if (r->type == PATTERN_ROW_TYPE_NOTE_ON) {
		strcpy(a_field_str, note_string[GET_NOTE_INDEX(r->col0)]);
		sprintf(b_field_str, "%c", octave_character[GET_OCTAVE_INDEX(r->col0)]);
		sprintf(c_field_str, "%02x", (unsigned) r->col1);
	}
	else if (r->type == PATTERN_ROW_TYPE_NOTE_OFF) {
		strcpy(a_field_str, "^^");
		strcpy(b_field_str, "^");
		strcpy(c_field_str, "--");
	}
	else if (r->type == PATTERN_ROW_TYPE_CONTROLLER) {
		sprintf(a_field_str, "%02x", (unsigned) r->col0);
		strcpy(b_field_str, "x");
		sprintf(c_field_str, "%02x", (unsigned) r->col1);
	}
	put_string(a_field_str, x + 1, y, a_field_palette);
	put_string(b_field_str, x + 3, y, b_field_palette);
	put_string(c_field_str, x + 5, y, c_field_palette);
	if (last)
		put_character('|', x + 7, y, border_palette);
}


void draw_pattern_track(pattern_track *t, u16 num_rows, u16 x, u16 y, u16 row_offset, u16 hilight_step, u16 last, u16 cursor_at_this_track) {
	u16 i, j, palette, hilighted_row, cursor_at_this_row;

	if (t == NULL)
		return;
	palette = LIGHT_GRAY_OVER_BLACK;
	for (i = 0, j = row_offset; (i < MAX_VISIBLE_ROWS) && (j < num_rows); i++, j++) {
		hilighted_row = 0;
		if (hilight_step > 0) {
			if ((j % hilight_step) == 0) {
				hilighted_row = 1;
				palette = WHITE_OVER_BLACK;
			}
			else
				palette = LIGHT_GRAY_OVER_BLACK;
		}
		cursor_at_this_row = cursor_at_this_track && (cursor_row == j);
		draw_pattern_row(&(t->data[j]), x, y + i, last, LIGHT_GRAY_OVER_BLACK, palette, cursor_at_this_row, hilighted_row);
	}
}


void draw_row_indeces(u16 x, u16 y, u16 row_offset, u16 num_rows, u16 palette) {
	u16 i, j;

	for (i = 0, j = row_offset; (i < MAX_VISIBLE_ROWS) && (j < num_rows); i++, j++) {
		sprintf(aux_str, "%02u", (unsigned) j);
		put_string(aux_str, x, y + i, palette);
	}
}


void draw_pattern(pattern *p, u16 pattern_index, u16 row_offset, u16 track_offset) {
	u16 i, j, x, cursor_at_this_track;
	u16 pat_palette, size_palette, hlstep_palette;
	u16 channel_palette;

	x = 8;
	for (i = 0, j = track_offset, x = 8; (i < MAX_VISIBLE_TRACKS) && (j < NUM_TRACKS); i++, j++, x += 7) {
		u16 last = (i < (MAX_VISIBLE_TRACKS - 1)) ? 0 : 1;
		sprintf(aux_str, "%02u", (unsigned) j);
		put_string(aux_str, 11 + (i * 7), 0, LIGHT_GRAY_OVER_BLACK);
		sprintf(aux_str, "%02u", (unsigned) current_song->midi_channel[j]);
		channel_palette = YELLOW_OVER_BLACK;
		if (editing_pattern_status == EDITING_PATTERN_STATUS_CHANNELS) {
			if (j == cursor_track)
				channel_palette = YELLOW_OVER_BLUE;
		}
		put_string(aux_str, 11 + (i * 7), 1, channel_palette);
		cursor_at_this_track = (j == cursor_track) ? 1 : 0;
		draw_pattern_track(&(p->tracks[j]), p->num_rows, x, 2, row_offset, p->hilight_step, last, cursor_at_this_track);
	}
	draw_row_indeces(6, 2, row_offset, p->num_rows, LIGHT_GRAY_OVER_BLACK);
	put_string("pat", 0, 2, LIGHT_GREEN_OVER_BLACK);
	sprintf(aux_str, "%02u", (unsigned) pattern_index);
	if ((status == STATUS_EDITING_PATTERN) && (editing_pattern_status == EDITING_PATTERN_STATUS_PAT))
		pat_palette = WHITE_OVER_BLUE;
	else
		pat_palette = LIGHT_GRAY_OVER_BLACK;
	put_string(aux_str, 0, 3, pat_palette);
	put_string("size", 0, 5, LIGHT_GREEN_OVER_BLACK);
	sprintf(aux_str, "%02u", (unsigned) p->num_rows);
	if ((status == STATUS_EDITING_PATTERN) && (editing_pattern_status == EDITING_PATTERN_STATUS_SIZE))
		size_palette = WHITE_OVER_BLUE;
	else
		size_palette = LIGHT_GRAY_OVER_BLACK;
	put_string(aux_str, 0, 6, size_palette);
	put_string("step", 0, 8, LIGHT_GREEN_OVER_BLACK);
	sprintf(aux_str, "%u", (unsigned) p->hilight_step);
	if ((status == STATUS_EDITING_PATTERN) && (editing_pattern_status == EDITING_PATTERN_STATUS_HLSTEP))
		hlstep_palette = WHITE_OVER_BLUE;
	else
		hlstep_palette = LIGHT_GRAY_OVER_BLACK;
	put_string(aux_str, 0, 9, hlstep_palette);
}


void draw_pattern_by_cursor(pattern *p, u16 pattern_index) {
	u16 last_display_track_offset, last_display_row_offset;

	/* track adjust */
	last_display_track_offset = display_track_offset + MAX_VISIBLE_TRACKS - 1;
	if (cursor_track < display_track_offset) {
		display_track_offset = cursor_track;
		last_display_track_offset = display_track_offset + MAX_VISIBLE_TRACKS - 1;
	}
	else if (cursor_track > last_display_track_offset) {
		last_display_track_offset = cursor_track;
		display_track_offset = last_display_track_offset - MAX_VISIBLE_TRACKS + 1;
	}
	/* row adjust */
	last_display_row_offset = display_row_offset + MAX_VISIBLE_ROWS - 1;
	if (cursor_row < display_row_offset) {
		display_row_offset = cursor_row;
		last_display_row_offset = display_row_offset + MAX_VISIBLE_ROWS - 1;
	}
	else if (cursor_row > last_display_row_offset) {
		last_display_row_offset = cursor_row;
		display_row_offset = last_display_row_offset - MAX_VISIBLE_ROWS + 1;
	}
	draw_pattern(p, pattern_index, display_row_offset, display_track_offset);
}


void draw_order(char *order, u16 offset) {
	u16 i, j;
	u16 x = 6;
	u16 palette;

	for (i = 0, j = offset; (i < MAX_VISIBLE_ORDERS) && (j < MAX_ORDERS); i++, j++) {
		if (status == STATUS_EDITING_ORDER) {
			if (j == cursor_order)
				palette = WHITE_OVER_BLUE;
			else
				palette = WHITE_OVER_BLACK;
		}
		else
			palette = LIGHT_GRAY_OVER_BLACK;
		if (order[j] >= MAX_PATTERNS)
			put_string("--", x, 14, palette);
		else {
			sprintf(aux_str, "%02d", (unsigned) order[j]);
			put_string(aux_str, x, 14, palette);
		}
		sprintf(aux_str, "%2d", (unsigned) j);
		put_string(aux_str, x, 15, LIGHT_GRAY_OVER_BLACK);
		x += 3;
	}
}


void draw_order_by_cursor(char *order) {
	u16 last_display_order_offset = display_order_offset + MAX_VISIBLE_ORDERS - 1;
	if (cursor_order < display_order_offset) {
		display_order_offset = cursor_order;
		last_display_order_offset = display_order_offset + MAX_VISIBLE_ORDERS - 1;
	}
	else if (cursor_order > last_display_order_offset) {
		last_display_order_offset = cursor_order;
		display_order_offset = last_display_order_offset - MAX_VISIBLE_ORDERS + 1;
	}
	draw_order(order, display_order_offset);
}


void draw_bpm(u16 bpm) {
	u16 palette;

	sprintf(aux_str, "%3d", bpm);
	if (status == STATUS_EDITING_BPM)
		palette = WHITE_OVER_BLUE;
	else
		palette = LIGHT_GRAY_OVER_BLACK;
	put_string(aux_str, 6, 16, palette);
}


void init_song_data(void) {
	u16 pi, ti, ri, i;

	memcpy(current_song->signature, "DINAMISE", 8);
	for (pi = 0; pi < MAX_PATTERNS; pi++) {
		pattern *pp = &(current_song->patterns[pi]);
		pp->num_rows = MAX_ROWS;
		pp->hilight_step = 4;
		for (ti = 0; ti < NUM_TRACKS; ti++) {
			pattern_track *pt = &(pp->tracks[ti]);
			for (ri = 0; ri < MAX_ROWS; ri++) {
				pattern_row *pr = &(pt->data[ri]);
				pr->type = PATTERN_ROW_TYPE_EMPTY;
				pr->col0 = 0;
				pr->col1 = 0;
			}
		}
	}
	for (i = 1; i < MAX_ORDERS; i++)
		current_song->order[i] = 0xFF;
	current_song->order[0] = 0;
	current_song->bpm = DEFAULT_BPM;
	for (i = 0; i < NUM_TRACKS; i++)
		current_song->midi_channel[i] = 1;
}


void file_option_new(void) {
	init_song_data();
	status_string = status_string_new_ok;
}


void file_option_load(void) {
	u32 i;
	u16 *to;
	u8 *from;

	if (memcmp(sram, "DINAMISE", 8) == 0) {
		from = (u8 *) sram;
		to = (u16 *) current_song;
		for (i = 0; i < 32768; i++) {
			u16 acc = *from;
			from++;
			acc |= ((u16) (*from)) << 8;
			*to = acc;
			to++;
			from++;
		}
		status_string = status_string_load_ok;
	}
	else
		status_string = status_string_load_error;
}


void file_option_save(void) {
	u16 *from;
	u8 *to;
	u32 i;

	from = (u16 *) current_song;
	to = (u8 *) sram;
	for (i = 0; i < 32768; i++) {
		*to = (u8) (*from & 0x00FF);
		to++;
		*to = (u8) (*from >> 8);
		to++;
		from++;
	}
	status_string = status_string_save_ok;
}


void draw_file_options(void) {
	u16 palette, i, x;
	file_option *fop;

	x = 5;
	fop = file_options;
	for (i = 0; i < NUM_FILE_OPTIONS; i++, fop++) {
		if (status == STATUS_FILE) {
			if (i == cursor_option)
				palette = WHITE_OVER_BLUE;
			else
				palette = WHITE_OVER_BLACK;
		}
		else
			palette = LIGHT_GRAY_OVER_BLACK;
		put_string(fop->label, x, 17, palette);
		x += strlen(fop->label) + 1;
	}
}


void show_main_screen(void) {
	wait_for_vsync();
	ERASE_SCREEN;
	put_string("trk", 7, 0, LIGHT_GREEN_OVER_BLACK);
	put_string("ch", 8, 1, LIGHT_GREEN_OVER_BLACK);
	put_string("order:", 0, 14, LIGHT_GRAY_OVER_BLACK);
	put_string("bpm:", 2, 16, LIGHT_GRAY_OVER_BLACK);
	put_string("file:", 0, 17, LIGHT_GRAY_OVER_BLACK);
	draw_pattern_by_cursor(&(current_song->patterns[current_pattern_index]), current_pattern_index);
	draw_order_by_cursor(current_song->order);
	draw_bpm(current_song->bpm);
	draw_file_options();
	put_string(status_string, 0, 19, LIGHT_GRAY_OVER_BLACK);
}


void show_playing_status(void) {
	if (status == STATUS_PLAYING) {
		sprintf(aux_str, "o:%02u r:%02u", (unsigned) playing_order_index, (unsigned) playing_row_index);
		put_string(aux_str, 21, 18, WHITE_OVER_BLACK);
		sprintf(aux_str, "%02u", (unsigned) midi_buffer_size);
		put_string(aux_str, 21, 19, WHITE_OVER_BLACK);
	}
	else {
		put_string("         ", 21, 18, WHITE_OVER_BLACK);
		put_string("  ", 21, 19, WHITE_OVER_BLACK);
	}
}


void increase_decrease_pattern_row_field(pattern *p, u16 cursor_track, u16 cursor_row, u16 cursor_field, u16 increase, u16 speed_up) {
	pattern_row *row;

	if (p == NULL)
		return;
	row = &(p->tracks[cursor_track].data[cursor_row]);
	if (cursor_field == PATTERN_ROW_FIELD_NOTE_CONTROLLER) {
		if (increase) {
			/* increase */
			if (speed_up)
				row->type = PATTERN_ROW_TYPE_NOTE_OFF;
			else {
				if (row->type == PATTERN_ROW_TYPE_EMPTY) {
					row->type = PATTERN_ROW_TYPE_NOTE_ON;
					row->col0 = 0;
					row->col1 = 0;
				}
				else if ((row->type == PATTERN_ROW_TYPE_NOTE_ON) || (row->type == PATTERN_ROW_TYPE_CONTROLLER)) {
					if (row->col0 < 127)
						row->col0++;
					else
						row->type = PATTERN_ROW_TYPE_NOTE_OFF;
				}
			}
		}
		else {
			/* decrease */
			if (speed_up)
				row->type = PATTERN_ROW_TYPE_EMPTY;
			else {
				if (row->type == PATTERN_ROW_TYPE_NOTE_OFF) {
					row->type = PATTERN_ROW_TYPE_NOTE_ON;
					row->col0 = 127;
					row->col1 = 0;
				}
				else if (row->type == PATTERN_ROW_TYPE_NOTE_ON) {
					if (row->col0 > 0)
						row->col0--;
					else
						row->type = PATTERN_ROW_TYPE_EMPTY;
				}
				else if (row->type == PATTERN_ROW_TYPE_CONTROLLER) {
					if (row->col0 > 0)
						row->col0--;
				}
			}
		}
	}
	else if (cursor_field == PATTERN_ROW_FIELD_OCTAVE_X) {
		if (increase) {
			if (row->type == PATTERN_ROW_TYPE_NOTE_ON) {
				u16 aux = GET_OCTAVE_INDEX(row->col0);
				if (aux == 10) {
					row->type = PATTERN_ROW_TYPE_CONTROLLER;
					row->col0 = 0;
					row->col1 = 0;
				}
				else
					row->col0 += 12;   /* octave up */
			}
		}
		else {
			/* decrease */
			if (row->type == PATTERN_ROW_TYPE_NOTE_ON) {
				if (row->col0 >= 12)
					row->col0 -= 12;   /* octave down */
			}
			else if (row->type == PATTERN_ROW_TYPE_CONTROLLER) {
				row->type = PATTERN_ROW_TYPE_NOTE_ON;
				row->col0 = 0;
				row->col1 = 0;
			}
		}
	}
	else if (cursor_field == PATTERN_ROW_FIELD_VELOCITY_VALUE) {
		if (increase) {
			s16 aux = (s16) row->col1;
			if (speed_up)
				aux += 0x10;
			else
				aux++;
			if (aux > 127)
				aux = 127;
			row->col1 = (unsigned char) aux;
		}
		else {
			/* decrease */
			s16 aux = (s16) row->col1;
			if (speed_up)
				aux -= 0x10;
			else
				aux--;
			if (aux < 0)
				aux = 0;
			row->col1 = (unsigned char) aux;
		}
	}
}


u16 status_editing_pattern_process_keys(void) {
	pattern *p = &(current_song->patterns[current_pattern_index]);

	if (key_pressed(BUTTON_START)) {
		status = STATUS_PLAYING;
		status_string = status_string_playing;
		start_playing();
		return 1;
	}
	if (key_pressed(BUTTON_SELECT)) {
		status = STATUS_EDITING_ORDER;
		return 1;
	}
	if (editing_pattern_status == EDITING_PATTERN_STATUS_PAT) {
		if (key_pressed(BUTTON_LEFT)) {
			editing_pattern_status = EDITING_PATTERN_STATUS_PATTERN;
			if (cursor_row >= p->num_rows)
				cursor_row = p->num_rows - 1;
			return 1;
		}
		else if (key_pressed(BUTTON_RIGHT)) {
			editing_pattern_status = EDITING_PATTERN_STATUS_SIZE;
			return 1;
		}
		else if (key_pressed(KEYPAD_UP)) {
			if (current_pattern_index < (MAX_PATTERNS - 1)) {
				current_pattern_index++;
				return 1;
			}
		}
		else if (key_pressed(KEYPAD_DOWN)) {
			if (current_pattern_index > 0) {
				current_pattern_index--;
				return 1;
			}
		}
	}
	else if (editing_pattern_status == EDITING_PATTERN_STATUS_SIZE) {
		if (key_pressed(BUTTON_LEFT)) {
			editing_pattern_status = EDITING_PATTERN_STATUS_PAT;
			return 1;
		}
		else if (key_pressed(BUTTON_RIGHT)) {
			editing_pattern_status = EDITING_PATTERN_STATUS_HLSTEP;
			return 1;
		}
		else if (key_pressed(KEYPAD_UP)) {
			if (p->num_rows < MAX_ROWS) {
				p->num_rows++;
				return 1;
			}
		}
		else if (key_pressed(KEYPAD_DOWN)) {
			if (p->num_rows > 1) {
				p->num_rows--;
				return 1;
			}
		}
	}
	else if (editing_pattern_status == EDITING_PATTERN_STATUS_HLSTEP) {
		if (key_pressed(BUTTON_LEFT)) {
			editing_pattern_status = EDITING_PATTERN_STATUS_SIZE;
			return 1;
		}
		else if (key_pressed(BUTTON_RIGHT)) {
			if (cursor_row >= p->num_rows)
				cursor_row = p->num_rows - 1;
			editing_pattern_status = EDITING_PATTERN_STATUS_CHANNELS;
			return 1;
		}
		else if (key_pressed(KEYPAD_UP)) {
			if ((p->hilight_step + 1) < MAX_ROWS) {
				p->hilight_step++;
				return 1;
			}
		}
		else if (key_pressed(KEYPAD_DOWN)) {
			if (p->hilight_step > 2) {
				p->hilight_step--;
				return 1;
			}
		}
	}
	else if (editing_pattern_status == EDITING_PATTERN_STATUS_CHANNELS) {
		if (key_pressed(BUTTON_LEFT)) {
			editing_pattern_status = EDITING_PATTERN_STATUS_HLSTEP;
			return 1;
		}
		else if (key_pressed(BUTTON_RIGHT)) {
			editing_pattern_status = EDITING_PATTERN_STATUS_PATTERN;
			return 1;
		}
		else if (key_pressed(KEYPAD_LEFT)) {
			if (cursor_track > 0) {
				cursor_track--;
				return 1;
			}
		}
		else if (key_pressed(KEYPAD_RIGHT)) {
			if (cursor_track < (NUM_TRACKS - 1)) {
				cursor_track++;
				return 1;
			}
		}
		else if (key_pressed(KEYPAD_UP)) {
			if (current_song->midi_channel[cursor_track] < 16) {
				current_song->midi_channel[cursor_track]++;
				return 1;
			}
		}
		else if (key_pressed(KEYPAD_DOWN)) {
			if (current_song->midi_channel[cursor_track] > 1) {
				current_song->midi_channel[cursor_track]--;
				return 1;
			}
		}
	}
	else if (editing_pattern_status == EDITING_PATTERN_STATUS_PATTERN) {
		if (key_pressed(BUTTON_LEFT)) {
			editing_pattern_status = EDITING_PATTERN_STATUS_CHANNELS;
			return 1;
		}
		else if (key_pressed(BUTTON_RIGHT)) {
			editing_pattern_status = EDITING_PATTERN_STATUS_PAT;
			return 1;
		}
		else if (key_pressed(KEYPAD_UP)) {
			if (PRESSED(BUTTON_A))
				increase_decrease_pattern_row_field(p, cursor_track, cursor_row, cursor_field, 1, 0);
			else {
				if (cursor_row > 0)
					cursor_row--;
			}
			return 1;
		}
		else if (key_pressed(KEYPAD_DOWN)) {
			if (PRESSED(BUTTON_A))
				increase_decrease_pattern_row_field(p, cursor_track, cursor_row, cursor_field, 0, 0);
			else {
				if (cursor_row < (p->num_rows - 1))
					cursor_row++;
			}
			return 1;
		}
		else if (key_pressed(KEYPAD_LEFT)) {
			if (PRESSED(BUTTON_A)) {
				increase_decrease_pattern_row_field(p, cursor_track, cursor_row, cursor_field, 0, 1);
				return 1;
			}
			else {
				if (cursor_field == PATTERN_ROW_FIELD_VELOCITY_VALUE) {
					cursor_field = PATTERN_ROW_FIELD_OCTAVE_X;
					return 1;
				}
				else if (cursor_field == PATTERN_ROW_FIELD_OCTAVE_X) {
					cursor_field = PATTERN_ROW_FIELD_NOTE_CONTROLLER;
					return 1;
				}
				else if (cursor_field == PATTERN_ROW_FIELD_NOTE_CONTROLLER) {
					if (cursor_track > 0) {
						cursor_track--;
						cursor_field = PATTERN_ROW_FIELD_VELOCITY_VALUE;
						return 1;
					}
				}
			}
		}
		else if (key_pressed(KEYPAD_RIGHT)) {
			if (PRESSED(BUTTON_A)) {
				increase_decrease_pattern_row_field(p, cursor_track, cursor_row, cursor_field, 1, 1);
				return 1;
			}
			else {
				if (cursor_field == PATTERN_ROW_FIELD_VELOCITY_VALUE) {
					if (cursor_track < (NUM_TRACKS - 1)) {
						cursor_track++;
						cursor_field = PATTERN_ROW_FIELD_NOTE_CONTROLLER;
						return 1;
					}
				}
				else if (cursor_field == PATTERN_ROW_FIELD_OCTAVE_X) {
					cursor_field = PATTERN_ROW_FIELD_VELOCITY_VALUE;
					return 1;
				}
				else if (cursor_field == PATTERN_ROW_FIELD_NOTE_CONTROLLER) {
					cursor_field = PATTERN_ROW_FIELD_OCTAVE_X;
					return 1;
				}
			}
		}
	}
	return 0;
}


u16 status_editing_order_process_keys(void) {
	char *current_order = &(current_song->order[cursor_order]);

	if (key_pressed(BUTTON_START)) {
		status = STATUS_PLAYING;
		status_string = status_string_playing;
		start_playing();
		return 1;
	}
	if (key_pressed(BUTTON_SELECT)) {
		status = STATUS_EDITING_BPM;
		return 1;
	}
	if (key_pressed(KEYPAD_LEFT)) {
		if (cursor_order > 0) {
			cursor_order--;
			return 1;
		}
	}
	else if (key_pressed(KEYPAD_RIGHT)) {
		if (cursor_order < (MAX_ORDERS - 1)) {
			cursor_order++;
			return 1;
		}
	}
	else if (key_pressed(KEYPAD_UP)) {
		if (*current_order == 0xFF) {
			*current_order = 0;
			return 1;
		}
		else if (*current_order < (MAX_PATTERNS - 1)) {
			(*current_order)++;
			return 1;
		}
	}
	else if (key_pressed(KEYPAD_DOWN)) {
		if (*current_order > 0) {
			(*current_order)--;
			return 1;
		}
		else if (*current_order == 0) {
			(*current_order) = 0xFF;
			return 1;
		}
	}
	return 0;
}


u16 status_editing_bpm_process_keys(void) {
	u16 *bpm = &(current_song->bpm);

	if (key_pressed(BUTTON_START)) {
		status = STATUS_PLAYING;
		status_string = status_string_playing;
		start_playing();
		return 1;
	}
	if (key_pressed(BUTTON_SELECT)) {
		status = STATUS_FILE;
		status_string = status_string_file;
		return 1;
	}
	if (key_pressed(KEYPAD_LEFT)) {
		if (*bpm > MIN_BPM) {
			(*bpm)--;
			return 1;
		}
	}
	else if (key_pressed(KEYPAD_RIGHT)) {
		if (*bpm < (MAX_BPM - 1)) {
			(*bpm)++;
			return 1;
		}
	}
	if (key_pressed(KEYPAD_DOWN)) {
		if (*bpm > (MIN_BPM + 10))
			(*bpm) -= 10;
		else
			*bpm = MIN_BPM;
		return 1;
	}
	else if (key_pressed(KEYPAD_UP)) {
		if (*bpm < (MAX_BPM - 11))
			(*bpm) += 10;
		else
			*bpm = MAX_BPM;
		return 1;
	}
	return 0;
}


u16 status_playing_process_keys(void) {
	if (key_pressed(BUTTON_START)) {
		status_string = status_string_editing;
		stop_playing();
		status = STATUS_EDITING_PATTERN;
		return 1;
	}
	return 0;
}


u16 status_file_process_keys(void) {
	if (key_pressed(BUTTON_START)) {
		status = STATUS_PLAYING;
		status_string = status_string_playing;
		start_playing();
		return 1;
	}
	if (key_pressed(BUTTON_SELECT)) {
		status = STATUS_EDITING_PATTERN;
		status_string = status_string_editing;
		return 1;
	}
	if (key_pressed(KEYPAD_LEFT)) {
		if (cursor_option > 0) {
			cursor_option--;
			return 1;
		}
	}
	else if (key_pressed(KEYPAD_RIGHT)) {
		if (cursor_option < (NUM_FILE_OPTIONS - 1)) {
			cursor_option++;
			return 1;
		}
	}
	else if (key_pressed(BUTTON_A)) {
		file_options[cursor_option].callback();
		return 1;
	}
	return 0;
}


int AgbMain(void) {
	u16 update_display = 1;
	/* set screen base block at 0x6000000,
	   character base block at 0x6004000 and
	   16 palettes of 16 colors each one */
	REG_DISPCNT = MODE0 | BG0_ENABLE;
	REG_BG0CNT = (1 << 2);
	/* define palettes and font data */
	define_palettes();
	dump_bitmap_font_to_tile_buffer(font, 2, 1);
	/* allocate room for song data (the song size MUST be under 64 Kbytes to store it in cartridge SRAM */
	current_song = (song *) malloc(sizeof(song));
	init_song_data();
	timer_0_start();
	while (1) {
		if (update_display) {
			show_main_screen();
			update_display = 0;
		}
		if (status == STATUS_EDITING_PATTERN)
			update_display = status_editing_pattern_process_keys();
		else if (status == STATUS_EDITING_ORDER)
			update_display = status_editing_order_process_keys();
		else if (status == STATUS_EDITING_BPM)
			update_display = status_editing_bpm_process_keys();
		else if (status == STATUS_FILE)
			update_display = status_file_process_keys();
		else if (status == STATUS_PLAYING) {
			poll_playing();
			show_playing_status();
			update_display = status_playing_process_keys();
		}
	}
}

